from dataclasses import asdict

import pytest

from idp_authentication import User, UserRole
from idp_authentication.conftest import TEST_PERMISSION, TEST_ROLE_1, TEST_ROLE_2


@pytest.mark.parametrize(
    "user,user_role,expected_role",
    [
        ("user", "test_role_3", "test_role_3"),
        ("user_2", "test_role_4", "test_role_4"),
    ],
    indirect=["user"],
)
def test_user_role_create(container, session, user, user_role, expected_role):
    process_user, user_as_dict = process_user_use_case(container, user, user_role)
    new_role = UserRole(role=user_role, app_entities_restrictions={}, user_id=user.id)
    user_as_dict["user_roles"] = new_role
    process_user.execute(user_as_dict)
    updated_user = session.query(User).filter(User.id == user.id).first()
    assert updated_user.user_roles[0].role == expected_role


@pytest.mark.parametrize(
    "user,user_role,expected_role",
    [
        ("user", TEST_ROLE_1, TEST_ROLE_1),
        ("user_2", TEST_ROLE_2, TEST_ROLE_2),
    ],
    indirect=["user"],
)
def test_user_role_update(container, session, user, user_role, expected_role):
    process_user, user_as_dict = process_user_use_case(container, user, user_role)
    process_user.execute(user_as_dict)
    updated_user = session.query(User).filter(User.id == user.id).first()

    assert updated_user.user_roles[0].role == expected_role


def test_process_user_deletes_user_roles_when_user_has_no_access_in_app(
    container, user
):
    # Arrange
    process_user, user_as_dict = process_user_use_case(
        container, user, TEST_ROLE_1, app_identifier="another_app"
    )

    # Act
    process_user.execute(user_as_dict)

    # Assert
    users_unit_of_work = container.users_module().users_unit_of_work()
    with users_unit_of_work:
        db_user = users_unit_of_work.user_repository.get_or_none(username=user.username)
        assert db_user.user_roles == []


def process_user_use_case(container, user, user_role, app_identifier="test"):
    process_user = container.users_module().use_cases().process_user_message_use_case()
    user_as_dict = asdict(user)
    user_as_dict["app_specific_configs"] = {
        app_identifier: {
            "default": {
                user_role: {
                    "app_entities_restrictions": {
                        "vehicle": [3],
                    },
                    "permission_restrictions": {TEST_PERMISSION: {"vehicle": [1, 2]}},
                }
            }
        }
    }
    return process_user, user_as_dict



from typing import Any

from idp_authentication.users.domain.ports.unit_of_work import UnitOfWorkPort


class BaseUnitOfWork(UnitOfWorkPort):
    def __init__(self, session_maker: Any):
        self.session_maker = session_maker

    def __enter__(self):
        self.session = self.session_maker()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Implicit commit in case everything goes right.
        # Otherwise, the rollback gets executed.
        if exc_type is None:
            self.commit()
        else:
            self.rollback()

        self.session_maker.remove()

    def commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()



from dependency_injector import containers, providers

from idp_authentication.users.adapters import unit_of_work
from idp_authentication.users.adapters.orm import start_mappers
from idp_authentication.users.domain import use_cases


class UsersModuleUseCasesDIContainer(containers.DeclarativeContainer):
    """Dependency injection container for use cases."""

    config = providers.Configuration(strict=True)
    users_unit_of_work = providers.Dependency()
    event_producer = providers.Dependency()
    get_user_use_case = providers.Factory(
        use_cases.GetUserUseCase,
        users_unit_of_work=users_unit_of_work,
    )
    create_or_update_user_use_case = providers.Factory(
        use_cases.CreateOrUpdateUserUseCase,
        users_unit_of_work=users_unit_of_work,
    )
    authorize_app_entity_records_use_case = providers.Factory(
        use_cases.AuthorizeEntityRecordsUseCase,
        users_unit_of_work=users_unit_of_work,
        roles=config.ROLES,
        app_entity_types=config.APP_ENTITY_TYPES,
    )
    get_allowed_record_identifiers_use_case = providers.Factory(
        use_cases.GetAllowedRecordIdentifiersUseCase,
        users_unit_of_work=users_unit_of_work,
        roles=config.ROLES,
        app_entity_types=config.APP_ENTITY_TYPES,
    )
    authorize_or_get_all_record_identifiers_use_case = providers.Factory(
        use_cases.AuthorizeOrGetAllAllowedRecordIdentifiersUseCase,
        users_unit_of_work=users_unit_of_work,
        roles=config.ROLES,
        app_entity_types=config.APP_ENTITY_TYPES,
        authorize_app_entity_records_use_case=authorize_app_entity_records_use_case,
        get_allowed_record_identifiers_use_case=get_allowed_record_identifiers_use_case,
    )
    get_users_with_access_to_app_entity_record_use_case = providers.Factory(
        use_cases.GetUsersWithAccessToAppEntityRecordUseCase,
        users_unit_of_work=users_unit_of_work,
        roles=config.ROLES,
        app_entity_types=config.APP_ENTITY_TYPES,
    )
    process_user_message_use_case = providers.Factory(
        use_cases.ProcessUserMessageUseCase,
        users_unit_of_work=users_unit_of_work,
        roles=config.ROLES,
        create_or_update_user_use_case=create_or_update_user_use_case,
        tenants=config.TENANTS,
        app_identifier=config.APP_IDENTIFIER,
    )
    send_app_entity_record_event_use_case = providers.Factory(
        use_cases.SendAppEntityRecordEventUseCase,
        event_producer=event_producer,
        app_identifier=config.APP_IDENTIFIER,
    )


class UsersModuleDIContainer(containers.DeclarativeContainer):
    """Dependency injection container for users module."""

    config = providers.Configuration(strict=True)
    database = providers.Dependency()
    event_producer = providers.Dependency()
    users_unit_of_work = providers.Factory(
        unit_of_work.UsersUnitOfWork,
        session_maker=database.provided.session_factory,
    )
    use_cases = providers.Container(
        UsersModuleUseCasesDIContainer,
        users_unit_of_work=users_unit_of_work,
        config=config,
        event_producer=event_producer,
    )
    start_mappers = providers.Callable(
        start_mappers,
        send_event_use_case=use_cases.send_app_entity_record_event_use_case,
    )



from idp_authentication.exceptions import UserNotFound
from idp_authentication.users.base_classes.base_use_case import UseCasePort
from idp_authentication.users.domain.entities.user import User
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort


class GetUserUseCase(UseCasePort):
    """
    Get user by idp_user_id.
    """

    def __init__(self, users_unit_of_work: UsersUnitOfWorkPort):
        self.users_unit_of_work = users_unit_of_work

    def execute(self, /, username: str) -> User:
        """Get user by username."""
        with self.users_unit_of_work as uow:
            user: User = uow.user_repository.get_first(username=username)
            if user is None:
                raise UserNotFound(username)

            return user



from typing import Any

from idp_authentication.enums import ChoiceEnum
from idp_authentication.users.base_classes.base_use_case import BaseUseCase


class GetUsersWithAccessToAppEntityRecordUseCase(BaseUseCase):
    def execute(
        self,
        app_entity_type: str,
        record_identifier: Any,
        roles: list[ChoiceEnum | str],
    ):
        self._verify_app_entity_type(app_entity_type)
        for role in roles:
            self._verify_role(role)

        with self.users_unit_of_work as uow:
            return uow.user_repository.get_users_with_access_to_records(
                app_entity_type=app_entity_type,
                record_identifier=record_identifier,
                roles=roles,
            )



import pytest

from idp_authentication import User
from idp_authentication.exceptions import UserNotFound


def test_get_user_use_case(container, user):
    get_user_use_case = container.users_module().use_cases().get_user_use_case()
    get_user = get_user_use_case.execute(username=user.username)
    assert isinstance(get_user, User)
    assert get_user.username == user.username


def test_get_user_use_case_raises_error_on_invalid_username(container):
    get_user_use_case = container.users_module().use_cases().get_user_use_case()
    with pytest.raises(
        UserNotFound, match="User with username=invalid_username not found!"
    ):
        get_user_use_case.execute(username="invalid_username")



from enum import Enum, EnumMeta


class ChoiceEnumMeta(EnumMeta):
    def __contains__(cls, item: int | str) -> bool:
        if isinstance(item, int):
            member_values = [v.value[0] for v in cls.__members__.values()]
        elif isinstance(item, str):
            item = item.lower()
            member_values = [v.value[1].lower() for v in cls.__members__.values()]
        else:
            member_values = cls.__members__.values()

        return item in member_values


class ChoiceEnum(Enum, metaclass=ChoiceEnumMeta):
    def __str__(self) -> str:
        return self.value[1]

    def __int__(self) -> int:
        return self.value[0]

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        elif isinstance(other, int):
            return int(self) == other
        return self is other

    @classmethod
    def get_by_value(cls, value):
        value_index = 0 if type(value) == int else 1
        return next(
            (v for v in cls.__members__.values() if v.value[value_index] == value), None
        )

    @classmethod
    def list_as(cls, item_type) -> list[int | str]:
        if item_type not in [int, str]:
            raise TypeError("Invalid item type")
        return list(map(item_type, cls))



from sqlalchemy import Boolean, Column, Date, DateTime, String

from idp_authentication.users.infrastructure.database.sqlalchemy.base import (
    Base,
    BaseModelMixin,
)


class UserModel(Base, BaseModelMixin):
    __tablename__ = "users"
    __table_args__ = {"extend_existing": True}
    username = Column(String(100), nullable=False, unique=True)
    email = Column(String(100), nullable=False, unique=True)
    first_name = Column(String(100), nullable=False)
    last_name = Column(String(100), nullable=False)
    is_active = Column(Boolean, nullable=False, default=True)
    is_superuser = Column(Boolean, nullable=False, default=False)
    is_staff = Column(Boolean, nullable=False, default=False)
    last_login = Column(DateTime, nullable=True)
    date_joined = Column(Date, nullable=True)



from idp_authentication.users.adapters.repositories.user_repository import (
    UserRepository,
)
from idp_authentication.users.adapters.repositories.user_role_repository import (
    UserRoleRepository,
)
from idp_authentication.users.base_classes.base_unit_of_work import BaseUnitOfWork
from idp_authentication.users.domain.ports.repository import (
    UserRepositoryPort,
    UserRoleRepositoryPort,
)
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort


class UsersUnitOfWork(BaseUnitOfWork, UsersUnitOfWorkPort):
    user_repository: UserRepositoryPort
    user_role_repository: UserRoleRepositoryPort

    def __enter__(self):
        super().__enter__()

        self.user_repository = UserRepository(self.session)
        self.user_role_repository = UserRoleRepository(self.session)

        return self
